import os
def parse_fid(f):
    #if the file has been processed through at least one step in the pipeline
    if "_" in f:
        parsed = str(f).split("_")
    #filename is FILENAME.bam i.e., hasn't been processed yet
    else:
        parsed = str(f).split(".")
    return parsed[0]

def is_correct_format(f):
    return str(f).count('.') >= 2

#checks that the file format follows our naming convenction
def check_file_format(directory):
    for f in os.listdir(directory):
        if not is_correct_format(f):
            raise ValueError(
                "All files saved in user-specified directory must follow the format 'FILENAME.bam'")

#check that all tools requested in steps are in the tools directory
def check_tools_exist(tools, steps, softwares):
    softwares = list(step for step in steps if step in softwares)
    for software in softwares:
        if not os.path.isdir(tools + "/" + software):
            raise ValueError(
                "User-specified 'tools' directory doesn't have a folder for " + software)

#creates subdirectories for all the requested steps within the specified output directory
def make_subdirectories(output, steps, slurm):
    subdirectories = {'remove_numts': [
        'fastqs', 'pileups', 'numt_removal_stor', 'counts']}
    for step in steps:
        if not os.path.isdir(output + "/" + step):
            os.makedirs(output + "/" + step)
            for sub in subdirectories[step]:
                os.makedirs(output + "/" + sub)
            #TODO: make subdirectories for parts within each step
    if slurm:
        os.makedirs(output + "/slurm")



from jinja2 import Template

#template for "import module" and "from module import function"
import_template = Template("{% for module in imports %}import {{ module }}\n{% endfor %}")

#template for directory and output paths
paths_template = Template(
    'STEPS = "./steps"\nPIPELINE_START = "{{ directory }}"\nPIPELINE_STORE = "{{ output }}"\nSLURM_DIR = "{{ output }}slurm"\n')

#template for a task without any requirements
task_template = Template(
    "class {{ task_name }}(luigi.Task):\n    id=luigi.Parameter()\n    def run(self):\n        subprocess.call([STEPS + \'/{{ job_name }}\', self.id, PIPELINE_START, PIPELINE_STORE + \'/{{ task_name }}\'])\n    def output(self):\n        return luigi.LocalTarget(PIPELINE_STORE + \'/{{ task_name }}/\' + \'{}_{{ file_name }}\'.format(self.id))")

#template for a slurm task without requirements
slurm_task_template = Template(
    "class {{ task_name }}(luigi.Task):\n    id=luigi.Parameter()\n    def run(self):\n        subprocess.call([STEPS + \'/submit_job.sh\', self.id, {{ job_name }}, SLURM_DIR, PIPELINE_START, PIPELINE_STORE + \'/{{ task_name }}\'])\n    def output(self):\n        return luigi.LocalTarget(PIPELINE_STORE + \'/{{ task_name }}/\' + \'{}_{{ file_name }}\'.format(self.id))")

#template for task with a required task
task_with_req_template = Template(
    "class {{ task_name }}(luigi.Task):\n    id=luigi.Parameter()\n    def require(self):\n        return {{ req_name }}(id=self.id)\n    def run(self):\n        subprocess.call([STEPS + \'/{{ job_name }}\', self.id, PIPELINE_START + \'/{{ req_name }}\', PIPELINE_STORE + \'/{{ task_name }}\'])\n    def output(self):\n        return luigi.LocalTarget(PIPELINE_STORE + \'/{{ task_name }}/\' + \'{}_{{ file_name }}\'.format(self.id))")

#slurm template for task with a required task
slurm_task_with_req_template = Template(
    "class {{ task_name }}(luigi.Task):\n    id=luigi.Parameter()\n    def require(self):\n        return {{ req_name }}(id=self.id)\n    def run(self):\n        subprocess.call([STEPS + \'/submit_job.sh\', self.id, {{ job_name }}, SLURM_DIR, PIPELINE_START + \'/{{ req_name }}\', PIPELINE_STORE + \'/{{ task_name }}\'])\n    def output(self):\n        return luigi.LocalTarget(PIPELINE_STORE + \'/{{ task_name }}/\' + \'{}_{{ file_name }}\'.format(self.id))")

#template for the wrapper task
wrapper_task_template = Template("class {{ task_name }}(luigi.WrapperTask):\n    _file = luigi.Parameter(default=\"\")\n    def requires(self):\n        if PIPELINE_START:\n            for f in os.listdir(PIPELINE_START):\n                if os.path.isfile(PIPELINE_START + f):\n                    {% for yield in yields %}yield {{ yield }}(id=util.parse_fid(f))\n                    {% endfor %}\n        else:\n            {% for yield in yields %}yield {{ yield }}(id=util.parse_fid(f))\n            {% endfor %}")



import argparse, sys, os
from pipeline_builder import PipelineBuilder
from pipeline_runner import PipelineRunner
class CommandLineParser():

    ALL_STEPS = ['extract_mito', 'split_gap', 'clipping', 'remove_numts', 'downsample',
                 'gatk', 'snpeff', 'annovar', 'haplogrep']
    DEPENDENCIES = {'snpeff': ['gatk'],
                'annovar': ['gatk'],
                'haplogrep': ['gatk'],
                'gatk': ['remove_numts'],
                'remove_numts': ['extract_mito', 'clip', 'split_gap']}

    def __init__(self, argv=sys.argv[1:]):
        self.__opts = self.parse_commands(argv)

    def parse_commands(self, argv=sys.argv[1:]):
        parser = self._build_parser()
        opts = parser.parse_args(argv)
        if not opts.directory:
            raise ValueError('Directory or file to be run on must be specified')
        return opts

    def build_and_run(self):
        pipeline_builder = PipelineBuilder()
        pipeline_builder.build_pipeline(slurm=self.__opts.slurm, tools=self.__opts.tools, directory=self.__opts.directory, steps=self.remove_steps(self.__opts.remove), output=self.__opts.output)
        #PipelineRunner.run('./pipeline.py', opts.workers)
        #if not opts.save:
        #    PipelineRunner.cleanup()
        
    def _build_parser(self):
        parser = argparse.ArgumentParser()
        #required arguments
        required_args = parser.add_argument_group('required arguments')
        required_args.add_argument('-d', '--directory', help="Path to the directory of files to be run", type=str)
        required_args.add_argument('-o', '--output', help="Output directory to save pipeline files")
        required_args.add_argument('-t', '--tools', help="Paths of locations of 3rd party packages")
        #optional arguments
        parser.add_argument('-s', '--save', help="Save files from the middle steps of pipeline instead of only the 3rd party software outputs", default=True, action='store_false')
        parser.add_argument('-l', '--slurm', help="Use slurm jobs to run each step", default=False, action='store_true')
        parser.add_argument('-r', '--remove', nargs='+', help="Steps to not run in this pipeline", default=None)
        #parser.add_argument('-c', '--config', help="Use the config file to specify software options", default=None)
        parser.add_argument('-w', '--workers', type=int, help="Number of workers to use to run the pipeline", default=1)

        return parser

    def remove_steps(self, steps):
        if not steps:
            return self.ALL_STEPS
        steps_to_use = self.ALL_STEPS
        for step in steps:
            if step in steps_to_use:
                steps_to_use.remove(step)
            else:
                raise ValueError("The requested step to remove doesn't exist in the pipeline")
        return steps_to_use



import sys, os

class PipelineRunner():

    @classmethod
    def run(cls, filename, opts):
        if not os.path.isfile(filename):
            raise FileNotFoundError("Pipeline file was not found. Ensure that the directory and output paths are correct")
        else:
            command = "PYTHONPATH=\'.\' luigi --module pipeline RunDataPipeline --workers " + opts.workers + " --local-scheduler"
            os.system(command)






import os, sys
from util import check_tools_exist, check_file_format, make_subdirectories
from templates import task_template, task_with_req_template, import_template, paths_template, wrapper_task_template, slurm_task_template, slurm_task_with_req_template

class PipelineBuilder():

    def __init__(self):
        self.softwares = ['gatk', 'snpeff', 'annovar', 'haplogrep']
        self.task_names = {'extract_mito': 'ExtractMito',
                            'split_gap': 'SplitGap',
                            'clipping': 'Clipping',
                            'remove_numts': 'RemoveNuMTs',
                            'downsample': 'Downsample',
                            'gatk': 'GATK',
                            'snpeff': 'SNPEFF',
                            'annovar': 'ANNOVAR',
                            'haplogrep': 'HAPLOGREP'}
        self.job_names = {'extract_mito': ['extract_mito.sh'],
                            'split_gap': ['split_gap.sh'],
                            'clipping': ['clipping.sh'],
                            'remove_numts': ['remove_numts.sh'],
                            'downsample': ['downsample.sh'],
                            'gatk': ['gatk.sh'],
                            'snpeff': ['snpeff.sh'],
                            'annovar': ['annovar.sh'],
                            'haplogrep': ['haplogrep.sh']}

    def build_pipeline(self, tools=None, slurm=False, directory=None, steps=['extract_mito', 'split_gap', 'clipping', 'remove_numts', 'downsample', 'gatk', 'snpeff', 'annovar', 'haplogrep'], output=None):
        if not directory:
            raise ValueError('Building the pipeline requires a file/directory to run on')
        if not output:
            raise ValueError('Pipeline requires a directory to store output files')
        if not tools:
            raise ValueError('Pipeline builder requires a directory that contains all the software tools')
        else:
            check_tools_exist(tools, steps, self.softwares)
            self.build_from_template(directory, output, steps, slurm)

    def build_from_template(self, directory, output, steps, slurm):
        check_file_format(directory)
        make_subdirectories(output, steps)
        with open('pipeline.py', 'w+') as pipeline:
            pipeline.write(import_template.render(imports=['util', 'luigi', 'subprocess', 'os', 'shutil']))
            pipeline.write(paths_template.render(directory=directory, output=output) + "\n\n")
            pipeline.write(wrapper_task_template.render(task_name="PipelineRunner", yields=list(self.task_names[step] for step in steps if step in self.softwares)) + "\n")
            prev_step = ""
           
            #write in the steps requested into the pipeline
            for step in steps:
                #if Complete Genomics data, i.e., did split gap then pipeline requires different scripts with shorter reads due to splitting into multiple reads at the gap
                if 'split_gap' in steps and step == 'remove_numts':
                    job_name = 'remove_numts_no_split_gap.sh'
                #step only is name of the step, not the name of the script
                else:
                    job_name = step + ".sh"
                pipeline.write(self.get_template(slurm, prev_step, job_name, step))
                if "gatk" in step or step not in self.softwares:
                    prev_step = self.task_names[step]
             
    def get_template(self, slurm, prev_step, job_name, step):
        #if slurm job requested
        if slurm:
            #first step in the pipeline has no 'require' function
            if prev_step == "":
                return slurm_task_template.render(task_name=self.task_names[step], job_name=job_name, file_name=step) + "\n\n"
            #if not the first step in the pipeline, require the previous step
            else:
                return slurm_task_with_req_template.render(task_name = self.task_names[step], req_name = prev_step, job_name = job_name, file_name = step) + "\n"
        else:
            #first step in the pipeline has no 'require' function
            if prev_step == "":
                return task_template.render(task_name=self.task_names[step], job_name=job_name, file_name=step) + "\n\n"
            #if not the first step in the pipeline, require the previous step
            else:
                return task_with_req_template.render(task_name=self.task_names[step], req_name=prev_step, job_name=job_name, file_name=step) + "\n"



from setuptools import setup

setup(
    name="mitopipeline",
    version="0.13",
    packages=['mitopipeline'],
    py_modules=['mitopipeline'],

    #add samtools, bwa, etc. here
    install_requires=[
        'jinja2'
    ],

    package_data={},

    #metadata
    author="Timothy Kuo",
    author_email="timykuo@gmail.com",
    description="Mitochondrial Genome Pipeline",
    license="MIT",
    keywords="genetics research data pipeline dependency management mitochondrial genome",
    url="https://github.com/timmykuo/mitopipeline",
    download_url="",

    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    project_urls={
        "Bug Tracker": "",
        "Documentation": "",
        "Source Code": "",
    }
)



import util
import luigi
import subprocess
import os
import shutil
STEPS = "./steps"
PIPELINE_START = "./"
PIPELINE_STORE = "./"
SLURM_DIR = "./slurm"

class PipelineRunner(luigi.WrapperTask):
    _file = luigi.Parameter(default="")
    def requires(self):
        if PIPELINE_START:
            for f in os.listdir(PIPELINE_START):
                if os.path.isfile(PIPELINE_START + f):
                    yield GATK(id=util.parse_fid(f))
                    yield SNPEFF(id=util.parse_fid(f))
                    yield ANNOVAR(id=util.parse_fid(f))
                    yield HAPLOGREP(id=util.parse_fid(f))
                    
        else:
            yield GATK(id=util.parse_fid(f))
            yield SNPEFF(id=util.parse_fid(f))
            yield ANNOVAR(id=util.parse_fid(f))
            yield HAPLOGREP(id=util.parse_fid(f))
            
class ExtractMito(luigi.Task):
    id=luigi.Parameter()
    def run(self):
        subprocess.call([STEPS + '/submit_job.sh', self.id, extract_mito.sh, SLURM_DIR, PIPELINE_START, PIPELINE_STORE + '/ExtractMito'])
    def output(self):
        return luigi.LocalTarget(PIPELINE_STORE + '/ExtractMito/' + '{}_extract_mito'.format(self.id))

class SplitGap(luigi.Task):
    id=luigi.Parameter()
    def require(self):
        return ExtractMito(id=self.id)
    def run(self):
        subprocess.call([STEPS + '/submit_job.sh', self.id, split_gap.sh, SLURM_DIR, PIPELINE_START + '/ExtractMito', PIPELINE_STORE + '/SplitGap'])
    def output(self):
        return luigi.LocalTarget(PIPELINE_STORE + '/SplitGap/' + '{}_split_gap'.format(self.id))
class Clipping(luigi.Task):
    id=luigi.Parameter()
    def require(self):
        return SplitGap(id=self.id)
    def run(self):
        subprocess.call([STEPS + '/submit_job.sh', self.id, clipping.sh, SLURM_DIR, PIPELINE_START + '/SplitGap', PIPELINE_STORE + '/Clipping'])
    def output(self):
        return luigi.LocalTarget(PIPELINE_STORE + '/Clipping/' + '{}_clipping'.format(self.id))
class RemoveNuMTs(luigi.Task):
    id=luigi.Parameter()
    def require(self):
        return Clipping(id=self.id)
    def run(self):
        subprocess.call([STEPS + '/submit_job.sh', self.id, remove_numts.sh, SLURM_DIR, PIPELINE_START + '/Clipping', PIPELINE_STORE + '/RemoveNuMTs'])
    def output(self):
        return luigi.LocalTarget(PIPELINE_STORE + '/RemoveNuMTs/' + '{}_remove_numts'.format(self.id))
class Downsample(luigi.Task):
    id=luigi.Parameter()
    def require(self):
        return RemoveNuMTs(id=self.id)
    def run(self):
        subprocess.call([STEPS + '/submit_job.sh', self.id, downsample.sh, SLURM_DIR, PIPELINE_START + '/RemoveNuMTs', PIPELINE_STORE + '/Downsample'])
    def output(self):
        return luigi.LocalTarget(PIPELINE_STORE + '/Downsample/' + '{}_downsample'.format(self.id))
class GATK(luigi.Task):
    id=luigi.Parameter()
    def require(self):
        return Downsample(id=self.id)
    def run(self):
        subprocess.call([STEPS + '/submit_job.sh', self.id, gatk.sh, SLURM_DIR, PIPELINE_START + '/Downsample', PIPELINE_STORE + '/GATK'])
    def output(self):
        return luigi.LocalTarget(PIPELINE_STORE + '/GATK/' + '{}_gatk'.format(self.id))
class SNPEFF(luigi.Task):
    id=luigi.Parameter()
    def require(self):
        return GATK(id=self.id)
    def run(self):
        subprocess.call([STEPS + '/submit_job.sh', self.id, snpeff.sh, SLURM_DIR, PIPELINE_START + '/GATK', PIPELINE_STORE + '/SNPEFF'])
    def output(self):
        return luigi.LocalTarget(PIPELINE_STORE + '/SNPEFF/' + '{}_snpeff'.format(self.id))
class ANNOVAR(luigi.Task):
    id=luigi.Parameter()
    def require(self):
        return GATK(id=self.id)
    def run(self):
        subprocess.call([STEPS + '/submit_job.sh', self.id, annovar.sh, SLURM_DIR, PIPELINE_START + '/GATK', PIPELINE_STORE + '/ANNOVAR'])
    def output(self):
        return luigi.LocalTarget(PIPELINE_STORE + '/ANNOVAR/' + '{}_annovar'.format(self.id))
class HAPLOGREP(luigi.Task):
    id=luigi.Parameter()
    def require(self):
        return GATK(id=self.id)
    def run(self):
        subprocess.call([STEPS + '/submit_job.sh', self.id, haplogrep.sh, SLURM_DIR, PIPELINE_START + '/GATK', PIPELINE_STORE + '/HAPLOGREP'])
    def output(self):
        return luigi.LocalTarget(PIPELINE_STORE + '/HAPLOGREP/' + '{}_haplogrep'.format(self.id))



import configparser

args = ['directory', 'file', 'output', 'steps', 'workers']

#required arguments
files = {'directory': None, 'file': None}
output = None

#optional arguments
steps = ['extract_mito', 'split_gap', 'clip', 'remove_numts', 'downsample',
         'gatk', 'heteroplasmy_analysis', 'gatk', 'snpeff', 'annovar', 'haplogrep']
workers = 1

def set_arg(arg, val):
    if arg not in args:
        raise ValueError('Input argument is not an available option')
    elif arg == "directory" or arg == "file":
        files.arg = val
    else:
        files.arg = val

